// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <cstring>
#include "xrm_ml_interface.h"
#include "xrm_utils.h"
#include "log_ama.h"

static constexpr uint32_t ML_COREID_CUID_OFFSET    = 8;
static constexpr uint32_t ML_COREID_GROUPID_OFFSET = 4;
typedef enum { ML_CORE_0 = 0, ML_CORE_1 = 1, ML_CORE_2 = 2, ML_CORE_3 = 3 } XRM_ML_GROUP_IDs;

/**
 * @brief Release the xrm resources which were allocated. Destroy the xrm 
 * context.
 * 
 * @param xrm_ml_ctx: The xrm ml context
 */
void xrm_ml_release(XrmMLContext* xrm_ml_ctx) {
  if (!xrm_ml_ctx) {
    return;
  }
  if (xrm_ml_ctx->ml_res_in_use) {
    /* Release the resource (still reserved) */
    xrmCuListReleaseV2(xrm_ml_ctx->xrm_ctx, &xrm_ml_ctx->ml_cu_list_res);
    xrm_ml_ctx->ml_res_in_use = 0;
  }
  xrmDestroyContext(xrm_ml_ctx->xrm_ctx);
}

/**
 * @brief Updates the group info for the ml core
 *
 * @param ml_cu_pool_prop The pool property which contains the cu / group info
 * @param ml_core_id The id of the core (-1 to let xrm decide, otherwise 0, 1, 2, 3)
 */
static int32_t ml_update_group_info(xrmCuPoolPropertyV2* ml_cu_pool_prop, int ml_core_id) {
  xrmCuListPropertyV2* ml_list_props   = &ml_cu_pool_prop->cuListProp;
  uint32_t             constraint_info = XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_NULL;
  int32_t              group_id        = 0;

  if (ml_core_id >= ML_CORE_0 && ml_core_id <= ML_CORE_3) {
    constraint_info = XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_REAL_LOGIC_GROUP_ID;
    group_id        = ml_core_id + ML_COREID_GROUPID_OFFSET;
  } else if (ml_core_id != -1) {
    LOG_AMA_ERROR(LOG_AMA_LAYER_XRM, "XRM received invalid core_id %d", ml_core_id);
    return XRM_ERROR;
  }

  xrmCuPropertyLogicGroupInfoV2* ml_logic_group_info = &ml_list_props->cuProps[0].logicGroupInfo;
  ml_logic_group_info->constraintInfo                = constraint_info;
  ml_logic_group_info->groupId                       = group_id;
  return XRM_SUCCESS;
}

/**
 * @brief Check & allocate the necessary xrm resources given the relevant 
 * ml parameters
 * 
 * @param xrm_ml_ctx Used to store information about resources this function
 * allocates so they can be safely released later
 * @param dev_index The device which will be queried
 * @param xrm_props Contains relevant information for load calculation
 * @return XRM_SUCCESS on success XRM_ERROR on error
 */
int xrm_ml_reserve(XrmMLContext* xrm_ml_ctx, int dev_index, XrmInterfaceProperties* xrm_props) {
  xrmCuPoolPropertyV2 ml_cu_pool_prop;
  int                 ret = XRM_ERROR;
  memset(&ml_cu_pool_prop, 0, sizeof(ml_cu_pool_prop));

  xrm_ml_ctx->dev_index = dev_index;

  /* Create XRM context */
  xrm_ml_ctx->xrm_ctx = xrmCreateContext(XRM_API_VERSION_1);
  if (xrm_ml_ctx->xrm_ctx == NULL) {
    return XRM_ERROR_CONNECT_FAIL;
  }

  /* Calculate ml load based on ml properties */
  ret = xrm_utils_load_calc(xrm_ml_ctx->xrm_ctx, xrm_props, NULL, 0, "ML", &xrm_ml_ctx->ml_load, 0);
  if (ret <= XRM_ERROR) {
    LOG_AMA_ERROR(LOG_AMA_LAYER_XRM, "ML load calculation failed %d", ret);
    return XRM_ERROR;
  }
  xrm_utils_update_pool_props("ML", xrm_ml_ctx->dev_index, xrm_ml_ctx->ml_load, &ml_cu_pool_prop, 0);
  if (ml_update_group_info(&ml_cu_pool_prop, xrm_ml_ctx->ml_core_id) == XRM_ERROR) {
    return XRM_ERROR;
  }
  ret                       = xrm_utils_reserve_resources(xrm_ml_ctx->xrm_ctx, &ml_cu_pool_prop, &xrm_ml_ctx->ml_cu_list_res);
  xrm_ml_ctx->ml_res_in_use = ret == XRM_SUCCESS ? 1 : 0;
  xrm_ml_ctx->dev_index     = xrm_ml_ctx->ml_cu_list_res.cuResources[0].deviceId;
  if (ret == XRM_SUCCESS) {
    xrm_ml_ctx->ml_core_id = xrm_ml_ctx->ml_cu_list_res.cuResources[0].cuId - ML_COREID_CUID_OFFSET;
    if (xrm_ml_ctx->ml_core_id < ML_CORE_0 || xrm_ml_ctx->ml_core_id > ML_CORE_3) {
      LOG_AMA_ERROR(LOG_AMA_LAYER_XRM, "XRM calculated invalid core_id %d!", xrm_ml_ctx->ml_core_id);
      return XRM_ERROR;
    }
  }
  return ret;
}
